import logging

logger = logging.getLogger(__name__)


from PyQt6.QtCore import QCoreApplication


# this function must eb named identical to QCoreApplication.translate
# otherwise lupdate doesnt recognize it
def translate(context, s) -> str:
    return QCoreApplication.translate(context, s)



import platform
from typing import Dict, Optional, Tuple

import bdkpython as bdk
import hwilib.commands as hwi_commands
from PyQt6.QtWidgets import QDialog, QMessageBox, QPushButton, QVBoxLayout

from bitcoin_usb.address_types import AddressType
from bitcoin_usb.udevwrapper import UDevWrapper

from .device import USBDevice, bdknetwork_to_chain
from .i18n import translate


def get_message_box(
    text: str, icon: QMessageBox.Icon = QMessageBox.Icon.Information, title: str = ""
) -> QMessageBox:
    # Create the text box
    msg_box = QMessageBox()
    msg_box.setIcon(icon)
    msg_box.setText(text)
    msg_box.setWindowTitle(title)

    # Add standard buttons
    msg_box.setStandardButtons(QMessageBox.StandardButton.Ok)

    return msg_box


class DeviceDialog(QDialog):
    def __init__(self, parent, devices, network):
        super().__init__(parent)
        self.setWindowTitle(self.tr("Select the detected device"))
        self.layout = QVBoxLayout(self)

        # Creating a button for each device
        for device in devices:
            button = QPushButton(f"{device['type']} - {device['model']}", self)
            button.clicked.connect(lambda *args, d=device: self.select_device(d))
            self.layout.addWidget(button)

        self.selected_device = None
        self.network = network

    def select_device(self, device):
        self.selected_device = device
        self.accept()

    def get_selected_device(self):
        return self.selected_device


class USBGui:
    def __init__(
        self,
        network: bdk.Network,
        allow_emulators_only_for_testnet_works: bool = True,
        autoselect_if_1_device=False,
        parent=None,
    ) -> None:
        self.autoselect_if_1_device = autoselect_if_1_device
        self.network = network
        self.parent = parent
        self.allow_emulators_only_for_testnet_works = allow_emulators_only_for_testnet_works

    def get_device(self) -> Dict:
        allow_emulators = True
        if self.allow_emulators_only_for_testnet_works:
            allow_emulators = self.network in [bdk.Network.REGTEST, bdk.Network.TESTNET, bdk.Network.SIGNET]

        devices = hwi_commands.enumerate(
            allow_emulators=allow_emulators, chain=bdknetwork_to_chain(self.network)
        )
        if not devices:
            get_message_box(
                translate("bitcoin_usb", "No USB devices found"),
                title=translate("bitcoin_usb", "USB Devices"),
            ).exec()
            return {}
        if len(devices) == 1 and self.autoselect_if_1_device:
            return devices[0]
        else:
            dialog = DeviceDialog(self.parent, devices, self.network)
            if dialog.exec():
                return dialog.get_selected_device()
            else:
                get_message_box(
                    translate("bitcoin_usb", "No device selected"),
                    title=translate("bitcoin_usb", "USB Devices"),
                ).exec()
        return {}

    def sign(self, psbt: bdk.PartiallySignedTransaction) -> Optional[bdk.PartiallySignedTransaction]:
        selected_device = self.get_device()
        if not selected_device:
            return None

        try:
            with USBDevice(selected_device, self.network) as dev:
                return dev.sign_psbt(psbt)
        except Exception as e:
            if not self.handle_exception_sign(e):
                raise

        return None

    def get_fingerprint_and_xpubs(self) -> Optional[Tuple[str, Dict[AddressType, str]]]:
        selected_device = self.get_device()
        if not selected_device:
            return None

        try:
            with USBDevice(selected_device, self.network) as dev:
                return dev.get_fingerprint(), dev.get_xpubs()
        except Exception as e:
            if not self.handle_exception_get_fingerprint_and_xpubs(e):
                raise
        return None

    def get_fingerprint_and_xpub(self, key_origin: str) -> Optional[Tuple[str, str]]:
        selected_device = self.get_device()
        if not selected_device:
            return None

        try:
            with USBDevice(selected_device, self.network) as dev:
                return dev.get_fingerprint(), dev.get_xpub(key_origin)
        except Exception as e:
            if not self.handle_exception_get_fingerprint_and_xpubs(e):
                raise
        return None

    def sign_message(self, message: str, bip32_path: str) -> Optional[str]:
        selected_device = self.get_device()
        if not selected_device:
            return None

        try:
            with USBDevice(selected_device, self.network) as dev:
                return dev.sign_message(message, bip32_path)
        except Exception as e:
            if not self.handle_exception_sign_message(e):
                raise
        return None

    def display_address(
        self,
        address_descriptor: str,
    ) -> Optional[str]:
        selected_device = self.get_device()
        if not selected_device:
            return None

        try:
            with USBDevice(selected_device, self.network) as dev:
                return dev.display_address(
                    address_descriptor=address_descriptor,
                )
        except Exception as e:
            if not self.handle_exception_display_address(e):
                raise
        return None

    def set_network(self, network: bdk.Network):
        self.network = network

    def handle_exception_get_fingerprint_and_xpubs(self, exception: Exception) -> bool:
        self.show_error_message(str(exception))
        return True

    def handle_exception_sign_message(self, exception: Exception) -> bool:
        self.show_error_message(str(exception))
        return True

    def handle_exception_display_address(self, exception: Exception) -> bool:
        self.show_error_message(str(exception))
        return True

    def handle_exception_sign(self, exception: Exception) -> bool:
        self.show_error_message(str(exception))
        return True

    def show_error_message(self, text: str):

        os_name = platform.system()

        if os_name == "Linux":
            self.show_error_message_linux(text)
        else:
            self.show_error_message(text)

    def show_error_message_linux(self, text: str):
        # Create the text box
        msg_box = QMessageBox()
        msg_box.setIcon(QMessageBox.Icon.Critical)
        msg_box.setText(text)
        msg_box.setWindowTitle(translate("bitcoin_usb", "Error"))

        # Add standard buttons
        msg_box.setStandardButtons(QMessageBox.StandardButton.Ok)

        msg_box.setInformativeText(
            translate(
                "bitcoin_usb",
                "USB errors can appear due to missing udev files. Do you want to install udev files now?",
            )
        )

        # Add standard buttons
        msg_box.setStandardButtons(QMessageBox.StandardButton.Cancel)

        # Add a custom button
        install_button = QPushButton(translate("bitcoin_usb", "Install udev files"))
        msg_box.addButton(install_button, QMessageBox.ButtonRole.ActionRole)
        install_button.clicked.connect(self.linux_cmd_install_udev_as_sudo)

        # Show the text box and wait for a response
        msg_box.exec()

    def linux_cmd_install_udev_as_sudo(self):
        UDevWrapper().linux_cmd_install_udev_as_sudo()
        get_message_box(
            translate("bitcoin_usb", "Please restart your computer for the changes to take effect."),
            QMessageBox.Icon.Information,
            translate("bitcoin_usb", "Restart computer"),
        ).exec()



import os
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Optional

from hwilib.udevinstaller import UDevInstaller, _resource_path

from bitcoin_usb.i18n import translate


class UDevWrapper:
    def find_filename_in_dir(self, filename: str, dir: Path) -> Optional[Path]:
        """
        Searches for a file in the specified directory that matches the filename exactly.
        Returns the Path object for the file if found, otherwise None.

        Args:
        filename (str): The exact filename to search for.
        dir (Path): The directory path to search in.

        Returns:
        Optional[Path]: The path to the file that matches the filename exactly, or None if no match is found.
        """
        if not dir.is_dir():
            return None

        # Iterate over all files in the directory
        for file in dir.iterdir():
            if file.is_file() and file.name == filename:
                return file

        # Return None if no file was found
        return None

    def get_udev_destination(
        self,
    ) -> Path:
        return Path("/etc/udev/rules.d/")

    def get_udev_source(self, absolute: bool) -> str:
        source_dir = "udev"
        return _resource_path(source_dir) if absolute else source_dir

    def linux_execute_sudo_script(self, script_content: str):
        terminals = ["konsole", "gnome-terminal", "xterm", "lxterminal", "xfce4-terminal"]

        # Create a temporary file to write the shell script
        with tempfile.NamedTemporaryFile(delete=False, mode="w", suffix=".sh") as temp_script:
            # Write commands to temp file
            temp_script.write("#!/bin/bash\n")
            temp_script.write(f"""echo '{translate("bitcoin_usb","Executing the script")}:'\n""")
            temp_script.write("echo '''\n" + script_content + "\n'''\n")
            temp_script.write(script_content)  # Add actual commands to be executed after confirmation
            temp_script_path = temp_script.name

        # Make the temporary script executable
        os.chmod(temp_script_path, 0o755)

        # Command to be executed with sudo
        full_cmd = f"bash {temp_script_path}"

        # Find a terminal emulator that is available and execute the script
        try:
            found_terminal = False
            for terminal in terminals:
                if shutil.which(terminal):
                    found_terminal = True
                    exec_cmd = [terminal, "-e", full_cmd]
                    if terminal == "konsole":
                        exec_cmd = [terminal, "-e", full_cmd]  # Removed --hold
                    elif terminal in ["gnome-terminal", "xfce4-terminal"]:
                        exec_cmd = [terminal, "-x", full_cmd]
                    subprocess.run(exec_cmd)
                    break
            if not found_terminal:
                print(translate("bitcoin_usb", "No suitable terminal emulator found."), file=sys.stderr)
                return False
        finally:
            # Optionally remove the script after execution
            os.unlink(temp_script_path)

        return True

    def _linux_cmd_install_udev(
        self,
    ) -> bool:
        "Needs root permission.  So actually execute it with sudo_linux_cmd_install_udev"
        source_dir = self.get_udev_source(absolute=False)
        destination = self.get_udev_destination()
        installer = UDevInstaller()
        print(f"Copies the rules from {self.get_udev_source(absolute=True)} to {destination}")
        return installer.install(source=str(source_dir), location=str(destination))

    def linux_cmd_install_udev_as_sudo(self, sleep=3):
        "Calls python as sudo and runs _linux_cmd_install_udev"
        python_executable_path = sys.executable
        script_content = f"""
                        #!/bin/bash
                        sudo {python_executable_path} {Path(__file__).absolute()} --install_linux_udev
                        sleep {sleep}
                        """
        self.linux_execute_sudo_script(script_content)

    def _all_files_present(self, source: Path, destination: Path, file_extension: str) -> bool:
        """
        Checks if all files with a specific file extension in the 'source' directory are present in the 'destination' directory.

        Args:
        source (Path): Path to the source directory.
        destination (Path): Path to the destination directory.
        file_ending (str): The file extension to filter by, e.g., '.rules'.

        Returns:
        bool: True if all files with the specified file extension from the source are present in the destination, False otherwise.
        """
        if not source.is_dir() or not destination.is_dir():
            return False  # Return False if either path is not a directory

        # Gather filenames with the specified file extension from source directory
        source_files = {
            file.name for file in source.iterdir() if file.is_file() and file.suffix == file_extension
        }

        # Gather filenames with the specified file extension from destination directory
        destination_files = {
            file.name for file in destination.iterdir() if file.is_file() and file.suffix == file_extension
        }

        # Check if all filtered files from the source are in the destination
        return source_files <= destination_files  # Subset check

    def all_udev_files_installed(self) -> bool:
        source = Path(self.get_udev_source(absolute=True))
        destination = self.get_udev_destination()

        return self._all_files_present(source, destination, file_extension="rule")


if __name__ == "__main__":
    # since
    import argparse

    parser = argparse.ArgumentParser(description="Can install the udev rules")
    parser.add_argument("--install_linux_udev", action="store_true")
    parser.add_argument("--test", action="store_true")

    args = parser.parse_args()

    udev_wrapper = UDevWrapper()
    if args.install_linux_udev:
        if udev_wrapper._linux_cmd_install_udev():
            print(translate("bitcoin_usb", "Successfully installed the udev rules"))
        else:
            print(translate("bitcoin_usb", "Could not install udev rules"))

        if udev_wrapper.all_udev_files_installed():
            print(translate("bitcoin_usb", "All udev files were successfully copied"))

    else:
        udev_wrapper.linux_cmd_install_udev_as_sudo()



from typing import Optional

import bdkpython as bdk
from PyQt6.QtWidgets import (
    QComboBox,
    QLineEdit,
    QMainWindow,
    QPushButton,
    QTabWidget,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from bitcoin_usb.gui import USBGui


class MainWindow(QMainWindow):
    def __init__(self, network: bdk.Network):
        super().__init__()
        self.usb = USBGui(network=network)

        main_widget = QWidget()
        main_widget_layout = QVBoxLayout(main_widget)
        self.setCentralWidget(main_widget)

        self.combo_network = QComboBox(self)
        for n in bdk.Network:
            self.combo_network.addItem(n.name, userData=n)
        self.combo_network.setCurrentText(network.name)
        main_widget_layout.addWidget(self.combo_network)

        # Create a tab widget and set it as the central widget
        tab_widget = QTabWidget(self)
        main_widget_layout.addWidget(tab_widget)

        # Tab 1: XPUBs
        xpubs_tab = QWidget()
        xpubs_layout = QVBoxLayout(xpubs_tab)
        self.button = QPushButton(self.tr("Get xpubs"), xpubs_tab)
        self.button.clicked.connect(self.on_button_clicked)
        xpubs_layout.addWidget(self.button)
        self.xpubs_text_edit = QTextEdit(xpubs_tab)
        self.xpubs_text_edit.setReadOnly(True)
        xpubs_layout.addWidget(self.xpubs_text_edit)
        tab_widget.addTab(xpubs_tab, self.tr("XPUBs"))

        # Tab 2: PSBT
        psbt_tab = QWidget()
        psbt_layout = QVBoxLayout(psbt_tab)
        self.psbt_text_edit = QTextEdit(psbt_tab)
        self.psbt_text_edit.setPlaceholderText(self.tr("Paste your PSBT in here"))
        psbt_layout.addWidget(self.psbt_text_edit)
        self.psbt_button = QPushButton(self.tr("Sign PSBT"), psbt_tab)
        self.psbt_button.clicked.connect(self.sign)
        psbt_layout.addWidget(self.psbt_button)
        tab_widget.addTab(psbt_tab, self.tr("PSBT"))

        # Tab 3: Message Signing
        message_tab = QWidget()
        message_layout = QVBoxLayout(message_tab)
        self.message_text_edit = QTextEdit(message_tab)
        self.message_text_edit.setPlaceholderText(self.tr("Paste your text to be signed"))
        message_layout.addWidget(self.message_text_edit)
        self.message_address_index_line_edit = QLineEdit(message_tab)
        self.message_address_index_line_edit.setText("m/84h/0h/0h/0/0")
        self.message_address_index_line_edit.setPlaceholderText(self.tr("Address index"))
        message_layout.addWidget(self.message_address_index_line_edit)
        self.sign_message_button = QPushButton(self.tr("Sign Message"), message_tab)
        self.sign_message_button.clicked.connect(self.sign_message)
        message_layout.addWidget(self.sign_message_button)
        tab_widget.addTab(message_tab, self.tr("Sign Message"))

        # Tab 4: Show Address
        address_tab = QWidget()
        address_tab.setLayout(QVBoxLayout())
        self.descriptor_text_edit = QTextEdit(address_tab)
        self.descriptor_text_edit.setPlaceholderText(self.tr("Paste your descriptor to be signed"))
        address_tab.layout().addWidget(self.descriptor_text_edit)
        self.display_address_button = QPushButton(self.tr("Display Address"), address_tab)
        self.display_address_button.clicked.connect(self.display_address)
        address_tab.layout().addWidget(self.display_address_button)
        tab_widget.addTab(address_tab, self.tr("Display Address"))

        # Initialize the network selection

        self.combo_network.currentIndexChanged.connect(
            lambda idx: self.usb.set_network(bdk.Network[self.combo_network.currentText()])
        )

    def display_address(self) -> Optional[str]:
        return self.usb.display_address(address_descriptor=self.descriptor_text_edit.toPlainText())

    def sign_message(self):
        signed_message = self.usb.sign_message(
            self.message_text_edit.toPlainText(), self.message_address_index_line_edit.text()
        )
        if signed_message:
            self.message_text_edit.setText(signed_message)

    def sign(self):
        psbt = bdk.PartiallySignedTransaction(self.psbt_text_edit.toPlainText())
        self.psbt_text_edit.setText("")
        signed_psbt = self.usb.sign(psbt)
        if signed_psbt:
            self.psbt_text_edit.setText(signed_psbt.serialize())

    def on_button_clicked(self):
        self.xpubs_text_edit.setText("")
        fingerprint_and_xpus = self.usb.get_fingerprint_and_xpubs()

        if not fingerprint_and_xpus:
            return
        fingerprint, xpubs = fingerprint_and_xpus

        if xpubs:
            txt = "\n".join(
                [
                    f"{str(k)}: [{k.key_origin(self.usb.network).replace('m/',f'{ fingerprint}/')}]  {v}"
                    for k, v in xpubs.items()
                ]
            )

            self.xpubs_text_edit.setText(txt)
